/*
 * This file is part of the TuxNES project codebase.
 *
 * Please see the README and COPYING files for more information regarding
 * this project.
 *
 * $Id: emu.c,v 1.14 2002/02/28 00:28:48 jim Exp $
 *
 * Description: This file contains the main() function of the emulator.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

/* dreamcast defines */
#undef HAVE_LIBZ
#undef HAVE_DC_REST
#define HAVE_DC_RENDNAME 
#define HAVE_DC_GETOPT
#define HAVE_DC_PAL

#ifndef HAVE_DC
#ifdef HAVE_FEATURES_H
#include <features.h>
#endif /* HAVE_FEATRES_H */

#include <ctype.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#ifdef HAVE_LINUX_JOYSTICK_H
#include <linux/joystick.h>
#endif /* HAVE_LINUX_JOYSTICK_H */
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef HAVE_LIBZ
#include <zlib.h>
#include "unzip.h"
#include "ziploader.h"
#endif
#else /* !HAVE_DC */
#include "dc-glue.h"
#include <kos.h>
#include "dc-draw.h"
#include "dc-qsort.h"
#define HAVE_DC_SAVE
#define HAVE_DC_REST
#define DC_TEST_CD_ACCESS
#endif /* !HAVE_DC */

#include "consts.h"
#include "gamegenie.h"
#include "globals.h"
#include "mapper.h"
#include "renderer.h"
#include "sound.h"
#include "palettes.h"
// #include "M6502.h"

u_char	*ROM_BASE;
u_char	*VROM_BASE;
char	*filename;
char	*homedir;
char	 savefile[1024];
char	*tuxnesdir;                /* buffer for $HOME/.tuxnes dir */
char    *basefilename;  /* base filename without the extensions */
unsigned int	ROM_PAGES;
unsigned int	VROM_PAGES;
unsigned int    ROM_MASK;
unsigned int    VROM_MASK;
unsigned int    VROM_MASK_1k;
char *sample_format_name = "8";
int	dirtyheader = 0;
int	disassemble = 0;
int	dolink = 0;
int	MAPPERNUMBER = -1;
int	SRAM_ENABLED;
int	gamegenie = 0;
int	irqflag = 0;
int	mapmirror = 0;
int	mapperoverride = 0;
int     ignorebadinstr = 0;
int	showheader = 0;
int	unisystem = 0;
int	verbose = 0;

/* joystick variables */
unsigned char	jsaxes[2] = {2, 2};
unsigned char	jsbuttons[2] = {2, 2};
int	 jsfd[2] = {-1, -1};
int	 jsversion[2] = {0x000800, 0x000800};
char	*jsdevice[2] = {0, 0};
#ifndef HAVE_DC
char	*rendname = "auto";
char	*soundname = "oss";
#else
char	*rendname = "dc-tile";
char	*soundname = "dc";
#endif

#define MAX_FILES 1024
#define MAX_FILENAME_LEN 256

struct entry {
	char *name;
	int size;
	int len;
} files[MAX_FILES];

int nfiles;

/* You need to free() the returned pointer */
void show_header(void);
char *select_game(int);
void scan_files(void);
void splash_loading(void *v);

#ifdef HAVE_LIBM
extern unsigned int *ntsc_palette(double hue, double tint);
double hue = 332.0;
double tint = 0.5;
#endif /* HAVE_LIBM */

static void	loadpal(char *);
static void	restoresavedgame(void);

void	InitMapperSubsystem(void);
void	InitMapperSubsystem_m6502(void);
char	MapperName[MAXMAPPER + 1][30];
void	quit(void);
void	START(void);
void	translate(int);

unsigned char *palremap = 0, *paldata = 0;
unsigned int *NES_palette = 0;
int monochrome = 0;

/****************************************************************************/

void
quit(void)
{
#ifndef HAVE_DC_REST
	int			fd;
  int			x;
#endif
	file_t fd;
	int x;
	int ret = 0;

	/* Timer ticks down at 50MHz / 64 per tick.  This is
	 * 781,250 ticks per second. */
	double delta = (renderer_data.basetime - timer_count(2)) / (double)TPS;
	double fps;

	/* Clean up char pointers */
#ifndef HAVE_DC_SAVE
	// Freed correctly?
	free (tuxnesdir);
	free (basefilename);
#endif

	/* Save ram */
	for (x = 0; ((long long *) (RAM + 0x6000))[x] == 0; x++);
	/* LOOKHERE Off-by-one? */
	/* FIXME bleah this is a mess b/c of previous dc-load lockup */
	if (x < 1023) {
		fd = fs_open(savefile, O_RDWR);
		if (fd == 0) {
			ret = 3;
		} else {
			if (fs_write (fd, RAM + 0x6000, 8192) != 8192) {
				ret = 3;
			} else {
				fs_close(fd);
				ret = 2;
			}
		}
	} else {
		ret = 1;
	}

  // Note: if we wait too long, dcload will lock
  // and save will not occur, hence the stupid ret code below
 
  if (ret == 1) printf("save ram empty\n");
  if (ret == 2) printf("saved ram to %s!\n", savefile);
  if (ret == 3)	printf("error saving ram to %s\n", savefile);
  fprintf(stderr,"attempting to exit()\n");
  eprintf("inst/cycles %d/%d (%d/%d)\n", total_inst, total_cycles, total_inst/60, total_cycles/60);
  eprintf("total_temp_ctr %d\n\n", total_temp_ctr);
  eprintf("frames: %d\n", total_frames);
  eprintf("time  : %d.%02d sec\n", (int)delta, (int)((delta-(int)delta)*100.0f));
  fps = total_frames / delta;
  eprintf("fps   : %d.%02d\n", (int)fps, (int)((fps-(int)fps)*100.0f));
  profile_print();

}

/****************************************************************************/

static void
restoresavedgame(void)
{
	file_t fd;
	if ((fd = fs_open (savefile, O_RDONLY)) != 0)
	{
		fs_read (fd, RAM + 0x6000, 8192);
		fs_close (fd);
	}
#ifndef HAVE_DC_REST
  int fd, result, x;
  char buffer[1024];
  struct stat statbuf;
  *savefile = 0;
  strncpy (buffer, homedir, 1010);
  if (*buffer != 0 && buffer[strlen (buffer) - 1] == '/')
    buffer[strlen (buffer) - 1] = 0;
  strcat (buffer, "/.tuxnes");
  result = stat (buffer, &statbuf);
  if (result == 0)
    {
      for (x = strlen (filename) - 1; x > 0 && filename[x - 1] != '/';
           x--);
      strcat (buffer, "/");
      strncat (buffer, filename + x, 1019 - strlen (buffer));
      strcat (buffer, ".sav");
      strcpy (savefile, buffer);
    }
  if (!*savefile)
    {
      strncpy (buffer, filename, 1013);
      savefile[1013] = 0;
      for (x = strlen (buffer); x >= 0 && buffer[x] != '/'; x--)
        buffer[x] = 0;
      strcat (buffer, "tuxnes.tmp");
      result = open (buffer, O_CREAT | O_RDWR, 0666 & ~mask);
      unlink (buffer);
      if (result < 0)
        {
          buffer[strlen (buffer) - 10] = 0;
          fprintf (stderr, "Warning: can not write to directory %s\n", buffer);
          fprintf (stderr,
                   "Create directory ~/.tuxnes if you want to save games there instead.\n");
        }
      else
        close (result);
      strncpy (savefile, filename, 1019);
      savefile[1019] = 0;
      strcat (savefile, ".sav");
    }
  if ((fd = open (savefile, O_RDWR, 0666 & ~mask)) >= 0)
    {
      read (fd, RAM + 0x6000, 8192);
      close (fd);
    }
  else
    {
      /* In this special case, there is a ~/.tuxnes directory but no save
         file in it.  If there is a save file in the romfile directory,
         copy it into the ~/.tuxnes directory. */
      strncpy (buffer, filename, 1019);
      buffer[1019] = 0;
      strcat (buffer, ".sav");
      if ((fd = open (buffer, O_RDWR, 0666 & ~mask)) >= 0)
        {
          read (fd, RAM + 0x6000, 8192);
          close (fd);
        }
      else
        {
          /* Finally, if no save file is found either place, look for a
             iNES-type save file. */
          strncpy (buffer, filename, 1023);
          buffer[1023] = 0;
          buffer[strlen (buffer) - 4] = 0;
          strcat (buffer, ".sav");
          fd = open (buffer, O_RDWR, 0666 & ~mask);
          read (fd, RAM + 0x6000, 8192);
          close (fd);
        }
    }
#endif /* HAVE_DC */
}

/****************************************************************************/

void
loadpal(char *palfile)
{
	/* FIXME We don't take into account appended ROM palette data (paldata). */
	/* Note: original tuxnes could stand some serious cleanup here. */
	if (unisystem && !palremap)
		NES_palette = palettes[sizeof (palettes) / sizeof (*palettes) - 1].data;
	else
		NES_palette = palettes[0].data;
}

/****************************************************************************/

/* Display ROM header bytes */
void show_header(void)
{
	int x;

	fprintf (stderr, "iNES header bytes:\n  0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F\n  ");

	for (x = 0; x < 16; x++)
		fprintf (stderr, "%02X  ", ROM[x]);
	fprintf (stderr, "\n ");

	for (x = 0; x < 16; x++) {
		if ((ROM[x] >= 0x20) && (ROM[x] <= 0x7F))
			fprintf (stderr, " %c  ", ROM[x]);
		else
			fprintf (stderr, "    ");
	}
	fprintf (stderr, "\n");
}

/* calc_[v]rom_mask: calculate how many address bits are used
 * for [V]ROM, based on number of [V]ROM pages. */
unsigned int calc_rom_mask(unsigned int rom_pages)
{
	unsigned int rom_mask;
	if (rom_pages <= 2)
		rom_mask = 1;
	else if (rom_pages <= 4)
		rom_mask = 3;
	else if (rom_pages <= 8)
		rom_mask = 7;
	else if (rom_pages <= 16)
		rom_mask = 15;
	else if (rom_pages <= 32)
		rom_mask = 31;
	else
		rom_mask = 63;
	return rom_mask;
}

unsigned int calc_vrom_mask(unsigned int vrom_pages)
{
	unsigned int vrom_mask;
	if (vrom_pages <= 2)
		vrom_mask = 1;
	else if (vrom_pages <= 4)
		vrom_mask = 3;
	else if (vrom_pages <= 8)
		vrom_mask = 7;
	else if (vrom_pages <= 16)
		vrom_mask = 15;
	else if (vrom_pages <= 32)
		vrom_mask = 31;
	else if (vrom_pages <= 64)
		vrom_mask = 63;
	else
		vrom_mask = 127;
	return vrom_mask;
}

void override_mirroring(int cmirror)
{
	switch(cmirror) {
		case 1:
//			if (verbose && !hvmirror)
//				fprintf (stderr, "Overriding default vertical mirroring.\n");
			hvmirror = 1;
			break;
		case 2:
//			if (verbose && hvmirror)
//				fprintf (stderr, "Overriding default horizontal mirroring.\n");
			hvmirror = 0;
			break;
		case 3:
//			if (verbose && !nomirror && !osmirror)
//				fprintf (stderr, "Overriding default mirroring.\n");
			osmirror = 1;
			break;
		case 4:
//			if (verbose && !nomirror)
//				fprintf (stderr, "Overriding default mirroring.\n");
			nomirror = 1;
			break;
//		default:
//			printf("Unknown mirroring override type; ignoring\n");
	}
}

int
main (int argc, char **argv)
{
#ifdef DYNAMIC_RECOMPILE
  int r;
#endif

  off_t size;
  int cmirror;
  int i, basestart, baseend; /* these are for working out the base filename */
  char *selected;
  kthread_t *thd;
  char *palfile = 0;	/* palette file */

#ifndef HAVE_DC_GETOPT
  char *ggcode = NULL;
  int ggret, parseret;

  /* for the Game Genie */
  int address, data, compare;
#endif

	file_t romfd;


#ifdef HAVE_DC
  code_cksum = cksum_add32((void *)0x8c010000, CODE_SIZE);
  eprintf("code cksum: %08x\n", code_cksum);
#endif

  /* set up the mapper arrays */
#ifdef DYNAMIC_RECOMPILE
  InitMapperSubsystem();
#else
  InitMapperSubsystem_m6502();
#endif

  /* initialize variables */
  verbose = 0;
  cmirror = 0;
  dolink = 0;
  disassemble = 0;
  gamegenie = 0;

#ifndef HAVE_DC
  sound_config.audiofile = NULL;
#endif




#define HAVE_DC_SOUND
#ifndef HAVE_DC_SOUND
  /* Select a sound sample format */
  {
    struct SampleFormat *match = 0;
    int partials = 0;
    int len;
    
    len = strlen (sample_format_name);
    for (sample_format = sample_formats; sample_format->name; sample_format++)
      if (!strcmp (sample_format->name, sample_format_name))
        {
	  break;
	}
      else if (!strncmp (sample_format->name, sample_format_name, len))
        {
	  match = sample_format;
	  partials++;
	}
    if (*sample_format_name && !sample_format->name)
      {
        char *tail;
	int sample_format_number;

	sample_format_number = strtoul (sample_format_name, &tail, 0);
	if (! *tail)
	  {
	    for (sample_format = sample_formats; sample_format->name; sample_format ++)
	      {
	        if (sample_format_number == sample_format->number) break;
	      }
	  }
      }
    if ((partials == 1) && !sample_format->name)
      sample_format = match;
    if (!sample_format->name)
      {
	if (partials)
	  fprintf (stderr, "%s: sound sample format name `%s' is ambiguous\n",
		   *argv, sample_format_name);
	else
	  fprintf (stderr, "%s: unrecognized sound sample format name `%s'\n",
		   *argv, sample_format_name);
	exit (2);
      }
  }
#else 
	// Force the sample_format to be 16-bit signed for now,
	// rather than looking for a match with the above code.
	// FIXME We can use the match code now if desired
	sample_format = &sample_formats[5];
#endif

  /* LOOKHERE */

	kos_init_all(THD_ENABLE | IRQ_ENABLE, ROMDISK_NONE);

	vid_init(DM_640x480, PM_RGB565);
	ta_init_defaults();
	draw_init();

	thd = thd_create(splash_loading, NULL);

	// Check for zero? 
	scan_files();

	// I'd think that if we destroy splash_loading while mid-frame,
	// the DC would choke.  Perhaps this can't happen.
	thd_destroy(thd);

	do {

	total_frames = 0;

#ifndef ROMDIR
#define ROMDIR "/pc/home/jim/build/tuxnes-dc/ng/"
#endif
#ifndef SAVDIR
#define SAVDIR "/pc/home/jim/build/tuxnes-dc/ng/save/"
#endif

	/* Kludge: I'm using savefile as a temporary buffer */
	strcpy(savefile, ROMDIR);
	if (!(selected = select_game(nfiles))) {
		printf("No game selected.\n");
		break;
	}
	strncat(savefile, selected, 1024 - strlen(savefile) - strlen(selected));
	free(selected);
	filename = savefile;

	ta_shutdown();

  fprintf(stderr,"filename: %s\n",filename);
  if ((romfd = fs_open (filename, O_RDONLY)) == 0) {
	  /* FIXME What if return value is negative? */
      fprintf(stderr,"failed to open %s\n",filename);
      exit(1);
  }
  fprintf(stderr,"romfd: %d\n",(int)romfd);
  /* hmm, not iso_lseek? */
  size = fs_seek (romfd, 0, SEEK_END);
  fs_seek (romfd, 0, SEEK_SET);   /* Get file size */
  fprintf(stderr,"size: %d\n",size);

  if (size < 0)
    {
      perror (filename);
      exit (1);
    }
  if (size == 0)
    {
      fprintf (stderr, "Unable to read %s (empty file)\n", filename);
      exit (1);
    }

	/* FIXME -pedantic says ROM is invalid lvalue */
	/* FIXME Why not allocate as much as we need? */
	ROM = (unsigned char *)malloc(0x200000);
	memset(ROM, 0, 0x200000);

	/* FIXME There needs to be a proper place for initialization of
	 * global variables to zero.  Many variables are set to zero via
	 * the BSS segment, so when we play a second game they are not
	 * reset to zero. */
	memset(vram, 0, 16384);

	fprintf(stderr, "try to read\n");
	if (fs_read(romfd, ROM, size) != size) {
		fprintf(stderr, "read failed\n");
      		perror(filename);
		exit(1);
	}
	fprintf(stderr, "read successful\n");
	fs_close(romfd);
	fprintf(stderr, "close successful\n");

	verbose=1;
  if (verbose)
    fprintf (stderr, "Rom size: %d\n", size);

 
  /*
    This initializes the base filename by taking the filename of the ROM and
    parsing of leading directory information as well as the trailing extension
    (.nes, .gz, .Z).
  */
  basestart = baseend = 0;
  for (i = 0; i < strlen(filename); i++)
    {
      if (filename[i] == '/')
        {
          basestart = i + 1;
        }
      else if (filename[i] == '.')
        {
          baseend = i - 1;
        }
    }

    if (! (basefilename = malloc(baseend - basestart + 1)))
      {
        perror (__FUNCTION__ ": malloc");
        exit (1);
      }

    for (i = 0; i <= (baseend - basestart); i++)
      {
        basefilename[i] = filename[basestart + i];
      }
    basefilename[baseend - basestart + 1] = '\0';

	/* Set up the save file name. */
	strcpy(savefile, SAVDIR);
	strncat(savefile, basefilename, 1024 - 4 - strlen(savefile));
	strcat(savefile, ".sav");
	printf("savefile: %s\n", savefile);

	/* Allocate memory */

	/* RAM = (unsigned char *)malloc(0x8000); */
	memset(RAM, 0x00, 0x8000);
/*	RAM[0x13e]=0x5a;  SMB1 level select */

  /* Choose sound engine */
  {
       int matches = 0; /* number of matches */
       struct Sounder *match = 0; /* first match */

       for (sounder = sounders; sounder->name; sounder++)
	    if (!strcmp (sounder->name, soundname)) {
		 match = sounder;
		 matches = 1;
		 break;
	    } else if (!strncmp (sounder->name, soundname, strlen(soundname))) {
		 match = sounder;
		 matches++;
	    }
       if (matches != 1) {
	    if (matches)
		 fprintf (stderr, "%s: sound engine name `%s' is ambiguous\n",
			  *argv, soundname);
	    else
		 fprintf (stderr, "%s: unrecognized sound engine name `%s'\n",
			  *argv, soundname);
	    exit (2);
       }
       sounder = match;
  }

  /* Initialize sound playback */
  if (sounder->InitAudio (argc, argv))
  {
       fprintf (stderr,
		"%s: warning: failed to initialize sound playback\n",
		*argv);
  }

  if (showheader) show_header();

  fprintf(stderr,"initializing mapper\n");
  /* Initialize ROM mapper */
  MAPTABLE[0] = RAM;
  MAPTABLE[1] = RAM;
  MAPTABLE[2] = RAM;
  MAPTABLE[4] = RAM;
  MAPTABLE[16] = RAM - 65536;   /* Wrap around accesses which index above FFFF */
  MAPTABLE[6] = MAPTABLE[7] = RAM;      /* Save RAM */
  MAPTABLE[3] = RAM + 0x2000;
  MAPTABLE[5] = RAM;            /* 3xxx and 5xxx really shouldn't be mapped to anything, but they're mapped to some unused ram just to prevent pagefaults by weird code that tries to access these areas. */
  SRAM_ENABLED = 1;
  if (!strncmp ("NES\x1a", (char *) ROM, 4))    /* .NES */
    {
      int used;

      ROM_BASE = ROM + 16;      /* 16 bytes for .nes header */

      /* check for battery-backed SRAM */
      if (!(ROM[6] & 2))
        SRAM_ENABLED = 0;

      /* check for a 512-byte trainer */
      if (ROM[6] & 4)
        {
          if (verbose)
            {
              fprintf (stderr, "512-byte trainer present\n");
            }
          memcpy (RAM + 0x7000, ROM_BASE, 512);
          ROM_BASE += 512;
        }

      /* figure out the mapper */
      if (!mapperoverride && dirtyheader)
        MAPPERNUMBER = ROM[6] >> 4;
      else if (!mapperoverride)
        {
          if ((ROM[7] & 0x0c) || ROM[8] | ROM[9]) {
            /* TuxNES doesn't override, but we will for DC */
            fprintf (stderr, "Warning: forcing 4-bit mapper\n");
            MAPPERNUMBER = ROM[6] >> 4;
          } else {
            MAPPERNUMBER = (ROM[6] >> 4) | (ROM[7] & 0xf0);
            /* there's actually a defined header bit past ROM[7]! */
            if ((ROM[10] & 0x10) && verbose)
              fprintf (stderr,
                       "Cartridge has bus conflicts (not emulated)\n");
            switch (ROM[7] & 0x03)
              {
              case 0x00:
                if (verbose)
                  fprintf (stderr, "NES/Famicom ROM\n");
                break;
              case 0x01:
                if (verbose)
                  fprintf (stderr, "VS UniSystem ROM\n");
                unisystem = 1;
                break;
              case 0x02:
                if (verbose)
                  fprintf (stderr, "PlayChoice-10 ROM\n");
                break;
              default:
                fprintf (stderr, "Warning: unrecognized system ID %d\n",
                         ROM[7] & 0x03);
            }
		  }
        }
      if (MAPPERNUMBER == 99)
        {
          unisystem = 1;
        }

      VROM_BASE = ROM_BASE + ROM[4] * 16384;
      if (ROM[5])
        memcpy (VRAM, VROM_BASE, 8192);
      ROM_PAGES = ROM[4];
      VROM_PAGES = ROM[5];

      /* calculate used space in image */
      used = 16                    /* header size */
	+ ((ROM[6] & 4) ? 512 : 0) /* trainer size */
	+ 16384 * ROM_PAGES        /* PRG-ROM pages */
	+ 8192 * VROM_PAGES;       /* CHR-ROM pages */

      /* handle palette remapping/redefinition tables */
      if (size == (used + 64)) {
	palremap = ROM + used;
      } else if (size == (used + 192)) {
	paldata = ROM + used;
      } else if ((size != used) && verbose) {
	fprintf (stderr, "File contains %d extra bytes\n",
		 size - used);
      }

      hvmirror = (ROM[6] & 1) ^ 1;
      nomirror = ROM[6] & 8;
    }
  else if (!strncmp ("FDS\x1a", (char *) ROM, 4))       /* .FDS */
    {
      fprintf (stderr,
            "Famicom Disk System (FDS) disk images aren't yet supported\n");
      exit (1);
    }
  else if (size == 40960)
    {
      /* No header, assume raw image */
      ROM_BASE = ROM;
      MAPPERNUMBER = 0;
      memcpy (VRAM, ROM + 32768, 8192);
      ROM_PAGES = 2;
      VROM_PAGES = 2;
      VROM_BASE = ROM + 32768;
      nomirror = 1;
      mapmirror = 0;
    }
  else
    {
      fprintf (stderr, "Unrecognized ROM file format\n");
      exit (1);
    }

  if (MAPPERNUMBER > MAXMAPPER || ((int *) MapperInit)[MAPPERNUMBER] == 0)
    {
      fprintf (stderr, "Unknown Mapper: %d (0x%02X)\n", MAPPERNUMBER,
               MAPPERNUMBER);
      exit (1);
    }
  if (verbose)
    {
#ifndef HAVE_DC
      fprintf (stderr, "Mapper %d (%s)  Program ROM %dk  Video ROM %dk\n",
               MAPPERNUMBER, MapperName[MAPPERNUMBER], ROM_PAGES * 16,
               VROM_PAGES * 8);
#else
      eprintf ("Mapper %d (%s)  Program ROM %dk  Video ROM %dk\n",
               MAPPERNUMBER, MapperName[MAPPERNUMBER], ROM_PAGES * 16,
               VROM_PAGES * 8);
#endif
    }

  /* How many ROM or VROM address bits are used? */
  ROM_MASK  = calc_rom_mask(ROM_PAGES);
  VROM_MASK = calc_vrom_mask(VROM_PAGES);

  /* VROM mask for 1k-sized pages */
  VROM_MASK_1k = (VROM_MASK << 3) | 7;

  MapperInit[MAPPERNUMBER]();

  if (cmirror && !mapmirror) override_mirroring(cmirror);

  /* What if mapmirror is set? */
  if (verbose && !mapmirror) {
      if      (nomirror) fprintf (stderr, "Using no mirroring.\n");
      else if (osmirror) fprintf (stderr, "Using one-screen mirroring.\n");
      else if (hvmirror) fprintf (stderr, "Using horizontal mirroring\n");
      else               fprintf (stderr, "Using vertical mirroring\n");
  }

  if (SRAM_ENABLED) restoresavedgame();

  if (!NES_palette) loadpal(palfile);

  /* This is a bit overkill--NES_palette should be a structure entry of
   * the same type as palettes[], and, if it is loaded from a file,
   * its name should be set to something like "file" or "filename.pal" */
  if (verbose)
    {
      int i;

      for (i = 0; i < (sizeof (palettes) / sizeof (*palettes)); i++)
        if (NES_palette == palettes[i].data)
          {
            fprintf (stderr, "Using built-in palette \"%s\"\n", palettes[i].name);
            break;
          }
    }

#ifndef HAVE_DC_GETOPT
  /* I'm not clear on what this is supposed to do */
  /* Looks like 64 byte suffix is appended to ROM to remap palette */
  /* Therefore, this doesn't have to do with getopt! */
  if (palremap) {
    unsigned int *new_palette;
    int pen;
    
    if (!(new_palette = (unsigned int *) malloc (64 * sizeof (*new_palette))))
      {
	fprintf(stderr, "Can't remap palette: ");
	fflush (stderr);
	perror ("malloc");
      }
    else
      {
	if (verbose)
	  fprintf (stderr, "Remapping palette\n");
	for (pen = 0; pen < 64; pen ++)
	  new_palette[pen] = (palremap[pen] <= 64)
	    ? NES_palette[palremap[pen]]
	    : NES_palette[pen];
	memcpy ((void *)NES_palette,
		(void *)new_palette,
		64 * sizeof(*NES_palette));
	free (new_palette);
      }
  }

  /* (Possibly) convert palette to monochrome */
  if (monochrome)
    {
      int pen;

      for (pen = 0; pen < 64; pen ++)
	{
	  unsigned long red, blue, green, gray;

	  red = (NES_palette[pen] & 0xFF0000) >> 16;
	  green = (NES_palette[pen] & 0xFF00) >> 8;
	  blue = NES_palette[pen] & 0xFF;
	  gray = 0.299 * red + 0.587 * green + 0.114 * blue;
	  if (gray > 0xFFU)
	    gray = 0xFFU;
	  NES_palette[pen] = (gray << 16) | (gray << 8) | gray;
	}
    }

  /* enter the Game Genie codes */
  if (gamegenie)
    {
      ggret = DecodeGameGenieCode (ggcode, &address, &data, &compare);
      if (ggret == GAME_GENIE_BAD_CODE)
        {
          fprintf (stderr, "invalid Game Genie code: %s\n", ggcode);
        }
      else if (ggret == GAME_GENIE_8_CHAR)
        {
          if (verbose)
            {
              fprintf (stderr, "Game Genie: address = %04X, data = %02X\n",
                       address, data);
              fprintf (stderr, "Game Genie: compare value = %02X\n", compare);
              fprintf (stderr, "Game Genie: value at %04X = %02X\n",
                       address, MAPTABLE[address >> 12][address]);
            }
          if (MAPTABLE[address >> 12][address] == compare)
            {
              MAPTABLE[address >> 12][address] = data;
              if (verbose)
                {
                  fprintf (stderr, "Game Genie: replacing...\n");
                  fprintf (stderr, "Game Genie: value at %04X = %02X\n",
                           address, MAPTABLE[address >> 12][address]);
                }
            }
        }
      else
        {
          if (verbose)
            {
              fprintf (stderr, "Game Genie: address = %04X, data = %02X\n",
                       address, data);
              fprintf (stderr, "Game Genie: replacing...\n");
            }
          MAPTABLE[address >> 12][address] = data;
          if (verbose)
            fprintf (stderr, "Game Genie: value at %04X = %02X\n", address,
                     MAPTABLE[address >> 12][address]);
        }
    }

#endif /* HAVE_DC_GETOPT */


  fprintf(stderr,"renderer: %s\n",rendname);
  /* Choose renderer */
  {
       int matches = 0; /* number of matches */
       struct Renderer *match = 0; /* first match */

       for (renderer = renderers; renderer->name; renderer++)
	    if (!strcmp (renderer->name, rendname)) {
		 match = renderer;
		 matches = 1;
		 break;
	    } else if (!strncmp (renderer->name, rendname, strlen(rendname))) {
		 match = renderer;
		 matches++;
	    }
       if (matches != 1) {
	    if (matches)
		 fprintf (stderr, "%s: renderer name `%s' is ambiguous\n",
			  *argv, rendname);
	    else
		 fprintf (stderr, "%s: unrecognized renderer name `%s'\n",
			  *argv, rendname);
	    exit (2);
       }
       renderer = match;
  }

  /* Initialize graphic display */
  if (renderer->InitDisplay (argc, argv))
  {
       fprintf (stderr,
		"%s: failed to initialize renderer `%s'\n",
		*argv, renderer->name);
       exit (1);
  }

  /* start the show */
  {
//	  extern M6502 *Rptr;
#ifdef DYNAMIC_RECOMPILE
	  STACKPTR = (int) STACK + 0xFF; 
	  translate (*((unsigned short *) (MAPTABLE[15] + 0xFFFC)));
/*	  START();                      */
#else
//	  M6502 R;
//	  Rptr = &R;
//	  R.IPeriod = 100;
/*	  Reset6502(&R);
	  Run6502(&R); */
#endif
#ifdef HAVE_DC
	  eprintf("starting cpu emulation\n");
#endif
/*RAM[0x6000] = 0x05;
RAM[0x6001] = 0x01;
R.IPeriod = 99999999; */

	  START();
  }

	  quit();
      sounder->ShutdownAudio();
	  free(ROM);
	  vid_init(DM_640x480, PM_RGB565);
	  ta_init_defaults();
	  draw_init();
	} while (1);
  
	for (i = 0; i < nfiles; i++) {
		free(files[i].name);
	}

  return(0);
}

int sort_entries(const void *s1, const void *s2)
{
	return strcmp(((struct entry *)s1)->name, ((struct entry *)s2)->name);
}

void scan_files(void)
{
	file_t fp;
	dirent_t *de;

	fp = fs_open(ROMDIR, O_RDONLY | O_DIR);
	if (!fp) { printf("Error opening %s\n", ROMDIR); return; }
	while ((de = fs_readdir(fp))) {
		int len;
		if (nfiles >= MAX_FILES) { printf("MAX_FILES exceeded!\n"); break; }
		/* printf("%s %d\n", de->name, de->size); */
		len = strlen(de->name);
		if (len > MAX_FILENAME_LEN) len = MAX_FILENAME_LEN;
		files[nfiles].name = malloc( len + 1 );
		files[nfiles].name[0] = '\0';
		strncat(files[nfiles].name, de->name, len);  // redundant, taken care of in malloc
		files[nfiles].size = de->size;
		files[nfiles].len = len;
		nfiles++;
	}
	fs_close(fp);

	printf("nfiles: %d\n", nfiles);

	qsort(files, nfiles, sizeof(struct entry), sort_entries);
}


/* Could return struct entry * */
char *select_game(int nfiles)
{
	/* Controller init is redundant */
	uint8 maple_controller = 0;
    cont_cond_t cond;
	unsigned int last_cond = 0;

	int frames = 0;
	float a = 0.0f, da = 0.03f;
	static int row = 0, top_row = 0;
	int i;
	int rpt_timer = 0; // 0.5 second

	char *selected = NULL;

	static char *rends[] = {"dc", "dc-tile", 0};
	static int rendnum = 0;
	char buf[1024];

	maple_controller = maple_first_controller();

#define MENU_X 85.0f
#define MENU_Y 85.0f
#define MENU_H_BORDER 10.0f
#define MENU_W_BORDER 10.0f
#define MENU_W (FONT_WIDTH * 38)
#define BOX_X  (MENU_X + MENU_W_BORDER)
#define BOX_Y  (MENU_Y + MENU_H_BORDER)
#define BOX_W  (MENU_W - MENU_W_BORDER * 2)
#define TEXT_X (BOX_X  + 5.0f)
#define TEXT_Y_SEP (4.0f)
#define TEXT_Y (BOX_Y  + (TEXT_Y_SEP / 2))
#define TEXT_W (BOX_W - 5.0f)
#define TEXT_H (FONT_HEIGHT)
#define BOX_H  (TEXT_H + TEXT_Y_SEP)
#define NUM_ROWS 10
#define MENU_H (BOX_H * NUM_ROWS + MENU_H_BORDER * 2)
#define REPEAT_TIME 4
#define DELAY_TIME  18

	do {
		/* Handle controller */
    	if (cont_get_cond(maple_controller, &cond)) {
//			printf("Controller 1 not present\n");
			break;
		}

	    if (!(cond.buttons & (CONT_START | CONT_B))) { 
			// Make sure it wasn't already pressed
			if ((last_cond & CONT_START)) break; 
		}
		if (!(cond.buttons & CONT_DPAD_DOWN)) {  
			if ((last_cond & CONT_DPAD_DOWN) || (rpt_timer <= 0)) { 
				if (top_row + row < nfiles - 1) {
					if (row < NUM_ROWS - 1) { row++; }
					else { top_row++; }
				}
				if (last_cond & CONT_DPAD_DOWN)  { rpt_timer = DELAY_TIME; }
				else { rpt_timer = REPEAT_TIME; }
			}
		}
		if (!(cond.buttons & CONT_DPAD_UP)) {  
			if ((last_cond & CONT_DPAD_UP) || (rpt_timer <= 0)) {
				if (top_row + row > 0) {
					if (row > 0) { row--; }
					else { top_row--; }
				}
				if (last_cond & CONT_DPAD_UP)  { rpt_timer = DELAY_TIME; }
				else { rpt_timer = REPEAT_TIME; }
			}
		}
		if (!(cond.buttons & CONT_A)) {  
			if ((last_cond & CONT_A)) { 
				printf("row %d: %s\n", row + top_row, files[row + top_row].name); 
				selected = malloc(files[row + top_row].len + 1);
				strcpy(selected, files[row + top_row].name); 
				break; 
			} /* FIXME error check for top_row+row<MAX_FILES (or nfiles) */
		}

		if (!(cond.buttons & CONT_Y)) {  
			if ((last_cond & CONT_Y)) { 
				rendnum++; if (!rends[rendnum]) rendnum=0;
			}
		}

		last_cond = cond.buttons;

		/* Render the frame */
		ta_begin_render();
		pvr_dummy_poly(TA_OPAQUE);
		ta_commit_eol();
		draw_box   ( MENU_X, MENU_Y, MENU_W, MENU_H, 0.3f, 0.1f, 0.6f, 0.4f);
		for (i = 0; i < NUM_ROWS; i++) {
			if (top_row + i >= nfiles) break;
			if (files[top_row + i].size >= 0) {
				draw_string( TEXT_X, TEXT_Y + (TEXT_H + TEXT_Y_SEP) * i, 2.0f, 1.0f, 1.0f, 1.0f, 1.0f, files[top_row + i].name, files[top_row + i].len);
			} else {
				draw_string( TEXT_X, TEXT_Y + (TEXT_H + TEXT_Y_SEP) * i, 2.0f, 1.0f, 1.0f, 1.0f, 0.3f, files[top_row + i].name, files[top_row + i].len);
			}
		}
		draw_box   ( BOX_X,  BOX_Y + BOX_H * row, BOX_W, BOX_H, a, 0.2f, 0.2f, 0.6f);
		draw_string(  35, 400, 2.0f, 1.0f, 0.8f, 1.0f, 0.8f, "A: select  B+START: quit", 24);
		// FIXME buffer overflow
		sprintf(buf, "Y: renderer (%s)", rends[rendnum]);
		draw_string(  35, 400 + TEXT_H + TEXT_Y_SEP, 2.0f, 1.0f, 0.8f, 1.0f, 0.8f, buf, strlen(buf));
		draw_string( 420, 400, 2.0f, 1.0f, 0.8f, 1.0f, 0.8f, "tuxnes-dc v0.2", 14);
		ta_commit_eol();
		ta_finish_frame();

		a += da;
		if (a >= 1.0f || a <= 0.0f) da = -da;

		if ((frames % (60 * 60)) == 0)
			printf(".");  // yuk kludge to prevent dc-load from hanging

		frames++;

		rpt_timer--;

	} while (1);

	rendname = rends[rendnum]; /* dangerous, use strncpy */

	return selected;
}	

void splash_loading(void *v)
{
	float a = 1.0f, da = -0.01f;
	char buf[10];
	for (;;) {
//		printf(" *** ");
//		thd_sleep(1000);
		ta_begin_render();
			pvr_dummy_poly(TA_OPAQUE);
			ta_commit_eol();
			draw_string( 450, 400, 2.0f, a, 0.8f, 1.0f, 0.8f, "Loading...", 10);
			// FIXME buffer overflow
			sprintf(buf, "%d", nfiles);
			draw_string( 450 + FONT_WIDTH * 10, 400, 2.0f, 1.0f, 0.8f, 0.8f, 1.0f, buf, strlen(buf));
			ta_commit_eol();
		ta_finish_frame();

		a += da;
		if (a >= 1.0f || a <= 0.3f) da = -da;
	}
}
