#include "dc-glue.h"
#include "globals.h" /* For vram */
#include "mapper.h" /* For VRAM -- yuck */

#define min(x, y) ( (x) < (y) ? (x) : (y) )

void draw_opaque_poly(float x1, float y1, float x2, float y2, float z, float r, float g, float b);
extern void draw_textured_pal_poly(uint32 txr_addr, float x, float y, float z, int bank);
extern void draw_textured_pal_poly_flip(uint32 txr_addr, float x, float y, float z, int hflip, int vflip, int bank);
extern void draw_textured_pal_poly_partial(uint32 txr_addr, float x, float y, float z, int clip_top, int clip_bot, int bank);
inline int tile_cache_is_valid(int tile, int color);
inline void tile_cache_validate(int tile, int color);
inline void tile_cache_invalidate(int tile);
inline void tile_cache_invalidate_all();
void tile_cache_initialize();
inline uint32 tile_cache_find(int tile, int color);
void render_one_tile(int tile, int color, int bank, int x, int y, int z);
void render_one_tile_flip(int tile, int color, int bank, int x, int y, int z, int hflip, int vflip);
void render_one_tile_partial(int tile, int color, int bank, int x, int y, int z, int clip_top, int clip_bot);
int load_one_tile(int tile, int color);

unsigned int tile_cache[512];
/* I could derive this mathematically, but let's be
 * more fleximable for now and store texture addresses per tile. */
uint32 tile_cache_txr_map[4][512];

#if 0
/* Test rendering of tiles */
void drawimage_tile(int endclock)
{
	int i, j;
	if (endclock == PBL) {
		for (j = 0; j < 16; j++) {
			for (i = 0; i < 32; i++) {
				render_one_tile(i + j*32, j & 3, i*10, j*10 + 40, 300);
			}
		}
	}
}
#else

void drawimage_tile(int endclock)
{
	int endclock_hpos, endclock_hpos0;
	static unsigned int lastclock = 0;
	unsigned int curclock = lastclock;
	unsigned int baseaddr = ((RAM[0x2000] & 0x10) << 8);  /* 0 or 0x1000 */
	unsigned int currentline = lastclock / HCYCLES;
	unsigned int hposition = lastclock % HCYCLES;
	unsigned int endline;

	int s;
	int hflip, vflip, behind;
	int spritebase = (RAM[0x2000] & 0x08) << 9;   /* 0x0 or 0x1000 */
	int spritesize = 8 << ((RAM[0x2000] & 0x20) >> 5);    /* 8 or 16 */
	int spritetile;

	unsigned int x, tile;
	uint8 tilecolor;
	int xpos, zpos = 300;
	static int ypos;

	int lines_left, lines_toadd;

	if (endclock > PBL) return;
	if (curclock > PBL) curclock = 0;
	if (endclock <= curclock) return;

	if (curclock == 0) {
		/* Begin new frame */
		ypos = 0;
		vline = vscrollreg >> 3;
		vscan = vscrollreg & 7;
		vwrap = 0;
		if (osmirror)
			scanpage = 0x2000;
		else if (nomirror)
			scanpage = 0x2000 + ((RAM[0x2000] & 3) << 10);
		else if (hvmirror == 0)
			scanpage = 0x2000 + ((RAM[0x2000] & 1) << 10);  /* v-mirror, h-layout */
		else if (hvmirror == 1)
			scanpage = 0x2000 + ((RAM[0x2000] & 2) << 9);   /* h-mirror, v-layout */
	}

	/* Adjust endclock so we only render complete lines. */
	endclock_hpos = endclock % HCYCLES;
	endclock_hpos0 = endclock - endclock_hpos;
	if (endclock_hpos < 85) {
		/* Ended in hblank; increase to end of hblank. */
		endclock = endclock_hpos0 + 84;
	} else {
		/* Ended mid-line; increase to end of next hblank. */
		endclock = endclock_hpos0 + HCYCLES + 84;
	}

	endline = endclock / HCYCLES;

	if (hposition >= 85) { 
		printf("error: hposition >= 85: %d\n", hposition); 
		*((volatile unsigned int *)3); /* assert */ 
	}

	/* Init */
	x = hscrollreg;
	if (osmirror)
		scanpage = 0x2000;
	else if (nomirror)
		scanpage = 0x2000 + (((RAM[0x2000] & 3) << 10) ^ (vwrap << 11));
	else if (hvmirror == 0)
		scanpage = 0x2000 + ((RAM[0x2000] & 1) << 10);      /* v-mirror, h-layout */
	else if (hvmirror == 1)
		scanpage = 0x2000 + (((RAM[0x2000] & 2) << 9) ^ (vwrap << 10));     /* h-mirror, v-layout */
	tile = VRAM[scanpage + ((x & 255) >> 3) + (vline << 5)];
	/* hposition should always be 84 (just before hblank end) at this point. */
	hposition = 85;
	tilecolor = VRAM[scanpage + 0x3C0 + ((x & 255) >> 5) + ((vline & 28) << 1)] >> ((vline & 2) << 1);
	/* This doesn't jibe with below */
	if (x & 16) tilecolor >>= 2;

	lines_left = endline - currentline;

	/* FIXME This bug is triggered by Zelda 2. */
	if (!lines_left) { printf("BUG: lines_left == 0 at line %d in %s\n", __LINE__, __FILE__); return; }

	while (lines_left)
	{
	     int t1, t2, t3;
	     t1 = scanpage + (vline << 5);
	     t2 = baseaddr + vscan;
	     t3 = t2 + 8;

		xpos = 32; 

		/* Force 8-pixel (tile) alignment by rounding down.  Also, adjust hposition 
		 * so we display an extra tile on the right, if alignment was necessary. */
		hposition -= x&7; 
		xpos      -= x&7; 
		x &= ~7;

		/* We may render partial tiles (top and/or bottom clipped).
		 * vscan tells us what line into the 8-line tile to begin (0-7).
		 * lines_toadd contains the number of lines to render during this pass.
		 * Whenever vscan reaches 8 we advance the tile row, so we can only render a maximum 
		 * of (8 - vscan) lines this pass---less if we reach the last line beforehand. */
		lines_toadd = min(8 - vscan, lines_left);

		/* FIXME Cleaner to rely on number of tiles rendered horizontally, rather than hposition */
		while (hposition < HCYCLES) {
			if ((RAM[0x2001] & 8)) {
				/* baseaddr (0 or 0x1000) refers to the first or second 256-tile block; each tile takes 16 bytes */
				/* note: vscan + lines_to_add <= 8 by definition */
				render_one_tile_partial(tile + (baseaddr >> 4), tilecolor & 3, PAL_BANK_TILE, xpos, ypos, zpos, vscan, vscan + lines_toadd);
			}

			x+=8;
			hposition+=8;
			xpos += 8;

			if (x == 256) {
				if ((!osmirror) && (nomirror || !hvmirror)) { 
					scanpage ^= 0x400;        // bit 8 of x -> bit 10 of addr
					t1 = scanpage + (vline << 5);
				} 
				x = 0;
			}

			tile = VRAM[t1 + (x >> 3)];

			if ((x & 0xf) == 0) {
				if ((x & 0x1f) == 0)
					tilecolor = VRAM[scanpage + 0x3C0 + (x >> 5) + ((vline & 28) << 1)] >> ((vline & 2) << 1);
				else
					tilecolor >>= 2;
			}
		}

		/* Sprites */
		/* We render whatever sprites are encountered in this row of tiles */
		/* We also render the second half of the 8x16 sprites at this time--
		 * so if sprite data changes halfway through, it will not be reflected. */
		if ((RAM[0x2001] & 16)) {
			int sx, sy, scolor;
			/* Check validity of s >= 0 expression */
			for (s = 63*4; s >= 0; s-=4) {
				sy = spriteram[s] + 1; /* add +1 sprite offset */
				sx = spriteram[s + 3];
				if (sy >= currentline && sy < currentline + 8 && sx < 256) {
						spritetile = spriteram[s + 1];
						behind = spriteram[s + 2] & 0x20;
						hflip = spriteram[s + 2] & 0x40;
						vflip = spriteram[s + 2] & 0x80;
						scolor = spriteram[s + 2] & 0x03;
						if (spritesize == 16) {
							/* 0x0000 for even 8x16 spritetiles; 0x1000 for odd */
							spritebase = (spritetile & 1) << 12;
							/* And so odd tiles become "even" */
							spritetile &= ~1;
							if (vflip) {
								render_one_tile_flip(spritetile + (spritebase >> 4), scolor, PAL_BANK_SPRITE, 32 + sx, sy + 8, behind ? 100 : 500, hflip, vflip);
								spritetile++;
							} else {
								render_one_tile_flip(spritetile + 1 + (spritebase >> 4), scolor, PAL_BANK_SPRITE, 32 + sx, sy + 8, behind ? 100 : 500, hflip, vflip);
							}
						}

						/* Tried subtracting s/4 from zpos, to effect sprite ordering 0-63, but it didn't work. */
						render_one_tile_flip(spritetile + (spritebase >> 4), scolor, PAL_BANK_SPRITE, 32 + sx, sy, behind ? 100 : 500, hflip, vflip);


				}
			}
		}

		x = hscrollreg;
		ypos += lines_toadd;
		lines_left -= lines_toadd;  /* see next comment */
		currentline += lines_toadd; /* these two are together redundant, but might be useful later */
		vscan += lines_toadd;
		/* Since lines_toadd <= 8 - vscan, vscan + lines_toadd <= 8. */
		if (vscan > 8) { printf("BUG: vscan %d > 8 at line %d\n", vscan, __LINE__); }
		if (vscan == 8) {
			/* New tile row. */
			vscan = 0;
			vline++;
			vline &= 31;
			if (vline == 30) {
				vline = 0;
				vwrap ^= 1;
			}
		}
		if (osmirror)
			scanpage = 0x2000;
		else if (nomirror)
			scanpage = 0x2000 + (((RAM[0x2000] & 3) << 10) ^ (vwrap << 11));
		else if (hvmirror == 0)
			scanpage = 0x2000 + ((RAM[0x2000] & 1) << 10);      /* v-mirror, h-layout */
		else if (hvmirror == 1)
			scanpage = 0x2000 + (((RAM[0x2000] & 2) << 9) ^ (vwrap << 10));     /* h-mirror, v-layout */
		tile = VRAM[scanpage + ((x & 255) >> 3) + (vline << 5)];
		tilecolor = VRAM[scanpage + 0x3C0 + ((x & 255) >> 5) + ((vline & 28) << 1)] >> ((vline & 2) << 1);
		if (x & 16) tilecolor >>= 2;

		hposition = 85;
	}

	lastclock = (endclock < PBL) ? endclock : 0;

}

#endif

void render_one_tile(int tile, int color, int bank, int x, int y, int z)
{
	if (!tile_cache_is_valid(tile, color)) {
		load_one_tile(tile, color);
	}

	/* Palette bank is 0 for tiles, 1 for sprites */
	draw_textured_pal_poly(tile_cache_find(tile, color), x, y, z, bank);
}

void render_one_tile_flip(int tile, int color, int bank, int x, int y, int z, int hflip, int vflip)
{
	if (!tile_cache_is_valid(tile, color)) {
		load_one_tile(tile, color);
	}

	/* Palette bank is 0 for tiles, 1 for sprites */
	draw_textured_pal_poly_flip(tile_cache_find(tile, color), x, y, z, hflip, vflip, bank);
}

void render_one_tile_partial(int tile, int color, int bank, int x, int y, int z, int clip_top, int clip_bot)
{
	if (!tile_cache_is_valid(tile, color)) {
		load_one_tile(tile, color);
	}

	/* Palette bank is 0 for tiles, 1 for sprites */
	draw_textured_pal_poly_partial(tile_cache_find(tile, color), x, y, z, clip_top, clip_bot, bank);
}


/* This should return a texture address */
/* Not unless render_one_tile calls it every time! */
int load_one_tile(int tile, int color)
{
	int i;
	uint8 byte1, byte2;
	uint8 tile_data[8*8/2];
	uint8 *tile_data_ptr = tile_data;
	unsigned int tileaddr = tile * 16; /* 8 bytes per color plane, 2 planes per tile */

	if (color & ~3) { printf("warning: color > 3: %d\n", color); color &= 3; }

	/* debug */
//	printf("load tile %d color %d\n", tile, color);

	for (i = 0; i < 8; i++) {
		byte1 = VRAM[tileaddr+i]; 
		byte2 = VRAM[tileaddr+i+8];

#if 1
		*tile_data_ptr++ = ((byte1 >> 7) & 0x01) | ((byte2 >> 6) & 0x02) | (color << 2) |
			              ((((byte1 >> 6) & 0x01) | ((byte2 >> 5) & 0x02) | (color << 2)) << 4);
		*tile_data_ptr++ = ((byte1 >> 5) & 0x01) | ((byte2 >> 4) & 0x02) | (color << 2) |
			              ((((byte1 >> 4) & 0x01) | ((byte2 >> 3) & 0x02) | (color << 2)) << 4);
		*tile_data_ptr++ = ((byte1 >> 3) & 0x01) | ((byte2 >> 2) & 0x02) | (color << 2) |
			              ((((byte1 >> 2) & 0x01) | ((byte2 >> 1) & 0x02) | (color << 2)) << 4);
		*tile_data_ptr++ = ((byte1 >> 1) & 0x01) | ((byte2 >> 0) & 0x02) | (color << 2) |
			              ((((byte1 >> 0) & 0x01) | ((byte2 << 1) & 0x02) | (color << 2)) << 4);
#else
		*tile_data_ptr++ = 0x33;
		*tile_data_ptr++ = 0x33;
		*tile_data_ptr++ = 0x77;
		*tile_data_ptr++ = 0x77;
#endif
	}

	txr_twiddle_copy_general(tile_data, tile_cache_find(tile, color), 8, 8, 4);
	tile_cache_validate(tile, color);
	return 0;
}

/* Wrapper functions for tile cache */

inline int tile_cache_is_valid(int tile, int color)
{
	return tile_cache[tile] & (1 << color);
}

inline void tile_cache_validate(int tile, int color)
{
	tile_cache[tile] |= (1 << color);
}

inline void tile_cache_invalidate(int tile)
{
	tile_cache[tile] = 0;
}

inline void tile_cache_invalidate_all()
{
	memset(tile_cache, 0, 512 * sizeof(unsigned int));
}

void tile_cache_initialize()
{
	int color, tile;
	for (tile = 0; tile < 512; tile++) {
		for (color = 0; color < 4; color++) {
			/* 4bpp 8x8 tiles */
			tile_cache_txr_map[color][tile] = ta_txr_allocate(8 * 8 / 2);
		}
	}
	tile_cache_invalidate_all();
}

/* Map tile/color combo to texture memory location. */
inline uint32 tile_cache_find(int tile, int color)
{
	return tile_cache_txr_map[color][tile];
}
